# Gregorian calendar in calendars package

# Copyright (c) 2022 Francesco Ricciardi
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name(s) of the copyright holders nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AS IS AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from fractions import Fraction


__author__ = "Francesco Ricciardi <francescor2010 at yahoo.it>"


def verify_value(num, den, min, max, min_excl, max_excl, strict):
    """Raised exceptions:
    - RuntimeError: if both min and min_excl, or max and max_excl are provided.
    - TypeError: if tuple argument for fraction is invalid or has wrong values
                 (0 denominator, NaN or similar)
    - ValueError: if fractional value does not observe condition(s)."""
    if min is not None and min_excl is not None:
        raise RuntimeError("Only one minimum value can be given.")
    if max is not None and max_excl is not None:
        raise RuntimeError("Only one maximum value can be given.")
    if strict:
        if den is None:
            if not isinstance(num, Fraction):
                raise TypeError("Value must be a Python Fraction.")
            den = 1
        else:
            if not isinstance(num, Fraction):
                raise TypeError("Numerator must be a Python Fraction.")
            if not isinstance(den, Fraction):
                raise TypeError("Denominator must be a Python Fraction.")
    try:
        value = Fraction(num, den)
    except TypeError as exc:
        raise TypeError("Invalid type in a fractional value.") from exc
    except (OverflowError, ValueError) as exc:
        raise TypeError("Invalid fractional value.") from exc
    if min is not None and value < min:
        raise ValueError(f"Value must be more than or equal to {min}")
    if min_excl is not None and value <= min_excl:
        raise ValueError(f"Value must be more than {min_excl}")
    if max is not None and value > max:
        raise ValueError(f"Value must be less than or equal to {max}")
    if max_excl is not None and value >= max_excl:
        raise ValueError(f"Value must be less than {max}")
    return value


def verify_fractional_value(fractional, min=None, max=None, min_excl=None, max_excl=None, strict=False):
    return verify_value(fractional, None, min, max, min_excl, max_excl, strict)


def verify_fractional_value_num_den(numerator, denominator, min=None, max=None, min_excl=None, max_excl=None, strict=False):
    return verify_value(numerator, denominator, min, max, min_excl, max_excl, strict)



# datetime2 package main file

# Copyright (c) 2011-2022 Francesco Ricciardi
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name(s) of the copyright holders nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AS IS AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

__author__ = "Francesco Ricciardi <francescor2010 at yahoo.it>"


import numbers
import time
from decimal import Decimal
from fractions import Fraction
from math import floor

from .common import verify_fractional_value, verify_fractional_value_num_den
from . import western, modern


##############################################################################
# OS dependent functions
#
def get_moment_complete():
    """Return local date and time as day_count, local time as day fraction, and,
    if possible, distance from UTC as fraction of a day."""
    try:
        moment_ns = (time.time_ns())  # time in ns from epoch; note epoch is platform dependent
    except AttributeError:
        moment_ns = int(time.time() * 1_000_000_000)  # time() returns a float in second
    # for the moment we are using time module's functions to get localtime
    # TODO: check if possible to implement something independent from time module, see e.g. tzlocal
    seconds, nanoseconds = divmod(moment_ns, 1_000_000_000)
    moment = time.localtime(seconds)
    year = moment.tm_year
    days_before_year = ((year - 1) * 365 + (year - 1) // 4 - (year - 1) // 100 + (year - 1) // 400)
    day_count = days_before_year + moment.tm_yday
    day_frac = (Fraction(moment.tm_hour, 24)   + Fraction(moment.tm_min, 1440) +
                Fraction(moment.tm_sec, 86400) + Fraction(nanoseconds, 86_400_000_000_000))
    utcoffset = Fraction(moment.tm_gmtoff, 86400)
    return day_count, day_frac, utcoffset


##############################################################################
#
# TimeDelta
#
##############################################################################


class TimeDelta:
    def __init__(self, numerator, denominator=None):
        if denominator is None:
            self._fractional_days = verify_fractional_value(numerator)
        else:
            self._fractional_days = verify_fractional_value_num_den(numerator, denominator)
        self._int_part = int(self._fractional_days)
        self._frac_part = self._fractional_days - self._int_part

    @property
    def fractional_days(self):
        return self._fractional_days

    @property
    def int_part(self):
        return self._int_part

    @property
    def frac_part(self):
        return self._frac_part

    def __repr__(self):
        return f"datetime2.{type(self).__name__}('{self.fractional_days!s}')"

    def __str__(self):
        if self._fractional_days == 0:
            return "0 days"
        int_string = f"{self.int_part!s} day{'s' if self._fractional_days != 1 and self._fractional_days != -1 else ''}" if self.int_part else ""
        frac_string = f"{self._frac_part!s} of a day" if self._frac_part else ""
        if int_string and frac_string:
            return f"{int_string} and {frac_string}"
        else:
            return int_string + frac_string

    # Math operators
    def int(self):
        return type(self)(self.int_part)

    def frac(self):
        return type(self)(self.frac_part)

    def __add__(self, other):
        if isinstance(other, TimeDelta):
            return type(self)(self.fractional_days + other.fractional_days)
        else:
            return NotImplemented

    def __pos__(self):
        return self

    def __neg__(self):
        return TimeDelta(-self.fractional_days)

    def __abs__(self):
        return self if self.fractional_days >= 0 else -self

    def __sub__(self, other):
        if isinstance(other, TimeDelta):
            return type(self)(self.fractional_days - other.fractional_days)
        else:
            return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (numbers.Real, Decimal)):
            try:
                return type(self)(self.fractional_days * Fraction(other))
            except (OverflowError, ValueError) as exc:
                reason = str(exc)
                if reason.startswith("cannot convert") and reason.endswith("to integer ratio"):
                    raise TypeError from exc
                else:
                    raise
        else:
            return NotImplemented

    __rmul__ = __mul__

    def __truediv__(self, other):
        if isinstance(other, TimeDelta):
            return self.fractional_days / other.fractional_days
        elif isinstance(other, (numbers.Real, Decimal)):
            try:
                return type(self)(self.fractional_days / Fraction(other))
            except (OverflowError, ValueError) as exc:
                reason = str(exc)
                if reason.startswith("cannot convert") and reason.endswith("to integer ratio"):
                    raise TypeError from exc
                else:
                    raise
        else:
            return NotImplemented

    def __floordiv__(self, other):
        if isinstance(other, TimeDelta):
            return self.fractional_days // other.fractional_days
        elif isinstance(other, (numbers.Real, Decimal)):
            try:
                return type(self)(self.fractional_days // Fraction(other))
            except (OverflowError, ValueError) as exc:
                reason = str(exc)
                if reason.startswith("cannot convert") and reason.endswith("to integer ratio"):
                    raise TypeError from exc
                else:
                    raise
        else:
            return NotImplemented

    def __mod__(self, other):
        if isinstance(other, TimeDelta):
            return self.fractional_days % other.fractional_days
        elif isinstance(other, (numbers.Real, Decimal)):
            try:
                return type(self)(self.fractional_days % Fraction(other))
            except (OverflowError, ValueError) as exc:
                reason = str(exc)
                if reason.startswith("cannot convert") and reason.endswith("to integer ratio"):
                    raise TypeError from exc
                else:
                    raise
        else:
            return NotImplemented

    def divmod(self, other):
        return self // other, self % other

    # boolean and test
    def __bool__(self):
        return self._fractional_days != 0

    def is_integer(self):
        return self._frac_part == 0

    # comparisons
    def __eq__(self, other):
        if isinstance(other, TimeDelta):
            return self._fractional_days == other.fractional_days
        elif hasattr(other, "fractional_days"):
            return NotImplemented
        else:
            return False

    def __ne__(self, other):
        if isinstance(other, TimeDelta):
            return self._fractional_days != other.fractional_days
        elif hasattr(other, "fractional_days"):
            return NotImplemented
        else:
            return True

    def __gt__(self, other):
        if isinstance(other, TimeDelta):
            return self._fractional_days > other.fractional_days
        elif hasattr(other, "fractional_days"):
            return NotImplemented
        else:
            raise TypeError(f"You cannot compare '{type(self)!s}' with '{type(other)!s}'.")

    def __ge__(self, other):
        if isinstance(other, TimeDelta):
            return self._fractional_days >= other.fractional_days
        elif hasattr(other, "fractional_days"):
            return NotImplemented
        else:
            raise TypeError(f"You cannot compare '{type(self)!s}' with '{type(other)!s}'.")

    def __lt__(self, other):
        if isinstance(other, TimeDelta):
            return self._fractional_days < other.fractional_days
        elif hasattr(other, "fractional_days"):
            return NotImplemented
        else:
            raise TypeError(f"You cannot compare '{type(self)!s}' with '{type(other)!s}'.")

    def __le__(self, other):
        if isinstance(other, TimeDelta):
            return self._fractional_days <= other.fractional_days
        elif hasattr(other, "fractional_days"):
            return NotImplemented
        else:
            raise TypeError(f"You cannot compare '{type(self)!s}' with '{type(other)!s}'.")

    # hash value
    def __hash__(self):
        return hash(self._fractional_days)


##############################################################################
#
# Date
#
##############################################################################


class Date:
    def __init__(self, day_count):
        # TODO: consider using the number hierarchy
        if isinstance(day_count, int):
            self._day_count = day_count
        else:
            raise TypeError("day_count argument for Date must be an integer.")

    @classmethod
    def today(cls):
        return cls(get_moment_complete()[0])

    @property
    def day_count(self):
        return self._day_count

    def __repr__(self):
        return f"datetime2.{type(self).__name__}({self.day_count})"

    def __str__(self):
        return f"R.D. {self.day_count}"

    def __add__(self, other):
        if isinstance(other, TimeDelta):
            if not other.is_integer():
                raise ValueError("Date object cannot be added to non integral TimeDelta.")
            return type(self)(self.day_count + int(other.fractional_days))   # note: fractiona days is a Fraction
        else:
            return NotImplemented

    __radd__ = __add__

    def __sub__(self, other):
        if isinstance(other, Date):
            return TimeDelta(self.day_count - other.day_count)
        elif isinstance(other, TimeDelta):
            if not other.is_integer():
                raise ValueError("Non integral TimeDelta cannot be subtracted from Date.")
            return type(self)(self.day_count - int(other.fractional_days))   # note: fractiona days is a Fraction
        else:
            return NotImplemented

    # Comparison operators
    def __eq__(self, other):
        if isinstance(other, Date):
            return self.day_count == other.day_count
        elif hasattr(other, "day_count"):
            return NotImplemented
        else:
            return False

    def __ne__(self, other):
        if isinstance(other, Date):
            return self.day_count != other.day_count
        elif hasattr(other, "day_count"):
            return NotImplemented
        else:
            return True

    def __gt__(self, other):
        if isinstance(other, Date):
            return self.day_count > other.day_count
        elif hasattr(other, "day_count"):
            return NotImplemented
        else:
            raise TypeError(f"You cannot compare '{type(self)!s}' with '{type(other)!s}'.")

    def __ge__(self, other):
        if isinstance(other, Date):
            return self.day_count >= other.day_count
        elif hasattr(other, "day_count"):
            return NotImplemented
        else:
            raise TypeError(f"You cannot compare '{type(self)!s}' with '{type(other)!s}'.")

    def __lt__(self, other):
        if isinstance(other, Date):
            return self.day_count < other.day_count
        elif hasattr(other, "day_count"):
            return NotImplemented
        else:
            raise TypeError(f"You cannot compare '{type(self)!s}' with '{type(other)!s}'.")

    def __le__(self, other):
        if isinstance(other, Date):
            return self.day_count <= other.day_count
        elif hasattr(other, "day_count"):
            return NotImplemented
        else:
            raise TypeError(f"You cannot compare '{type(self)!s}' with '{type(other)!s}'.")

    # hash value
    def __hash__(self):
        return hash(self._day_count)

    @classmethod
    def register_new_calendar(cls, attribute_name, calendar_class):
        if not isinstance(attribute_name, str) or not attribute_name.isidentifier():
            raise ValueError(f"Invalid calendar attribute name: {attribute_name}.")
        if hasattr(cls, attribute_name):
            raise AttributeError(f"Calendar attribute already existing: {attribute_name}.")
        if not hasattr(calendar_class, "from_rata_die"):
            raise TypeError("Calendar class does not have method from_rata_die.")
        if not hasattr(calendar_class, "to_rata_die"):
            raise TypeError("Calendar class does not have method to_rata_die.")

        class ModifiedClass(type):
            def __call__(klass, *args, **kwargs):
                calendar_obj = super().__call__(*args, **kwargs)
                date_obj = cls(calendar_obj.to_rata_die())
                setattr(date_obj, attribute_name, calendar_obj)
                return date_obj

        # Create the modified calendar class
        new_class_name = f"{calendar_class.__name__}In{cls.__name__}"
        modified_calendar_class = ModifiedClass(new_class_name, (calendar_class,), {})

        class CalendarAttribute:
            # This class implements a context dependent attribute
            def __init__(self, attribute_name, modified_calendar_class):
                self.attribute_name = attribute_name
                self.modified_calendar_class = modified_calendar_class

            def __get__(self, instance, owner):
                if instance is None:
                    return self.modified_calendar_class
                else:
                    assert self.attribute_name not in instance.__dict__
                    date_obj = self.modified_calendar_class.from_rata_die(instance.day_count)
                    calendar_obj = getattr(date_obj, self.attribute_name)
                    setattr(instance, self.attribute_name, calendar_obj)
                    return calendar_obj

        setattr(cls, attribute_name, CalendarAttribute(attribute_name, modified_calendar_class))


##############################################################################
# Register current calendars
#
Date.register_new_calendar("gregorian", western.GregorianCalendar)
Date.register_new_calendar("iso", modern.IsoCalendar)


##############################################################################
#
# Time
#
##############################################################################


class Time:
    def __init__(self, numerator, denominator=None, *, utcoffset=None):
        if denominator is None:
            self._day_frac = verify_fractional_value(numerator, min=0, max_excl=1)
        else:
            self._day_frac = verify_fractional_value_num_den(numerator, denominator, min=0, max_excl=1)
        if utcoffset is None:
            # naive instance
            self._utcoffset = None
        else:
            # aware instance
            self._utcoffset = verify_fractional_value(utcoffset, min=-1, max=1)

    @classmethod
    def now(cls, utcoffset=None):
        current_moment = get_moment_complete()
        if utcoffset is None:
            return cls(current_moment[1], utcoffset=current_moment[2])
        else:
            valid_utcoffset = verify_fractional_value(utcoffset, min=-1, max=1)
            delta = current_moment[2] - valid_utcoffset
            day_frac_temp = current_moment[1] - delta + 2  # +2 needed to avoid underruns
            new_day_frac = day_frac_temp - int(day_frac_temp)  # as so we eliminate the +2 above
            return cls(new_day_frac, utcoffset=valid_utcoffset)

    @classmethod
    def localnow(cls):
        current_moment = get_moment_complete()
        return cls(current_moment[1])

    @classmethod
    def utcnow(cls):
        current_moment = get_moment_complete()
        utcnow = current_moment[1] - current_moment[2]
        if utcnow < 0:
            utcnow += 1
        elif utcnow >= 1:
            utcnow -= 1
        return cls(utcnow)

    @property
    def day_frac(self):
        return self._day_frac

    @property
    def utcoffset(self):
        return self._utcoffset

    def __repr__(self):
        if self.utcoffset is None:
            return f"datetime2.{type(self).__name__}('{self.day_frac!s}')"
        else:
            return f"datetime2.{type(self).__name__}('{self.day_frac!s}', utcoffset='{self.utcoffset!s}')"

    def __str__(self):
        if self.utcoffset is None:
            return f"{self.day_frac!s} of a day"
        else:
            return f"{self.day_frac!s} of a day, {self.utcoffset!s} of a day from UTC"

    # Math operators
    def __add__(self, other):
        if isinstance(other, TimeDelta):
            total = self.day_frac + other.fractional_days
            return type(self)(total - floor(total), utcoffset=self.utcoffset)
        else:
            return NotImplemented

    __radd__ = __add__

    def __sub__(self, other):
        if isinstance(other, Time):
            if self.utcoffset:
                if other.utcoffset is None:
                    raise ValueError("You cannot mix naive and aware instances.")
                self_utc = self.day_frac - self.utcoffset
                other_utc = other.day_frac -  other.utcoffset
                delta = self_utc - other_utc
            else:
                if other.utcoffset is not None:
                    raise ValueError("You cannot mix naive and aware instances.")
                delta = self.day_frac - other.day_frac
            if delta <= Fraction(-1, 2):
                delta += 1
                while delta <= Fraction(-1, 2):
                    delta += 1
            elif delta > Fraction(1, 2):
                delta -= 1
                while delta > Fraction(1, 2):
                    delta -= 1
            return TimeDelta(delta)
        elif isinstance(other, TimeDelta):
            total = self.day_frac - other.fractional_days
            return type(self)(total - floor(total), utcoffset=self.utcoffset)
        else:
            return NotImplemented

    # Comparison operators
    def __eq__(self, other):
        if isinstance(other, Time):
            if self.utcoffset is None:
                if other.utcoffset is None:
                    return self.day_frac == other.day_frac
                else:
                    return False
            else:
                if other.utcoffset is not None:
                    return self.day_frac - self.utcoffset == other.day_frac - other.utcoffset
                else:
                    return False
        elif hasattr(other, "day_frac") and hasattr(other, "utcoffset"):
            return NotImplemented
        else:
            return False

    def __ne__(self, other):
        if isinstance(other, Time):
            if self.utcoffset is None:
                if other.utcoffset is None:
                    return self.day_frac != other.day_frac
                else:
                    return True
            else:
                if other.utcoffset is not None:
                    return self.day_frac - self.utcoffset != other.day_frac - other.utcoffset
                else:
                    return True
        elif hasattr(other, "day_frac") and hasattr(other, "utcoffset"):
            return NotImplemented
        else:
            return True

    def __gt__(self, other):
        if isinstance(other, Time):
            if self.utcoffset is None:
                if other.utcoffset is None:
                    return self.day_frac > other.day_frac
                else:
                    raise TypeError("You cannot compare a naive Time instance with an aware one.")
            else:
                if other.utcoffset is not None:
                    return self.day_frac - self.utcoffset > other.day_frac - other.utcoffset
                else:
                    raise TypeError("You cannot compare an aware Time instance with a naive one.")
        elif hasattr(other, "day_frac") and hasattr(other, "utcoffset"):
            return NotImplemented
        else:
            raise TypeError(f"You cannot compare '{type(self)!s}' with '{type(other)!s}'.")

    def __ge__(self, other):
        if isinstance(other, Time):
            if self.utcoffset is None:
                if other.utcoffset is None:
                    return self.day_frac >= other.day_frac
                else:
                    raise TypeError("You cannot compare a naive Time instance with an aware one.")
            else:
                if other.utcoffset is not None:
                    return self.day_frac - self.utcoffset >= other.day_frac - other.utcoffset
                else:
                    raise TypeError("You cannot compare an aware Time instance with a naive one.")
        elif hasattr(other, "day_frac") and hasattr(other, "utcoffset"):
            return NotImplemented
        else:
            raise TypeError(f"You cannot compare '{type(self)!s}' with '{type(other)!s}'.")

    def __lt__(self, other):
        if isinstance(other, Time):
            if self.utcoffset is None:
                if other.utcoffset is None:
                    return self.day_frac < other.day_frac
                else:
                    raise TypeError("You cannot compare a naive Time instance with an aware one.")
            else:
                if other.utcoffset is not None:
                    return self.day_frac - self.utcoffset < other.day_frac - other.utcoffset
                else:
                    raise TypeError("You cannot compare an aware Time instance with a naive one.")
        elif hasattr(other, "day_frac") and hasattr(other, "utcoffset"):
            return NotImplemented
        else:
            raise TypeError(f"You cannot compare '{type(self)!s}' with '{type(other)!s}'.")

    def __le__(self, other):
        if isinstance(other, Time):
            if self.utcoffset is None:
                if other.utcoffset is None:
                    return self.day_frac <= other.day_frac
                else:
                    raise TypeError("You cannot compare a naive Time instance with an aware one.")
            else:
                if other.utcoffset is not None:
                    return self.day_frac - self.utcoffset <= other.day_frac - other.utcoffset
                else:
                    raise TypeError("You cannot compare an aware Time instance with a naive one.")
        elif hasattr(other, "day_frac") and hasattr(other, "utcoffset"):
            return NotImplemented
        else:
            raise TypeError(f"You cannot compare '{type(self)!s}' with '{type(other)!s}'.")

    # hash value
    def __hash__(self):
        if self.utcoffset is None:
            return hash((self.day_frac, None))
        else:
            return hash(self.day_frac - self.utcoffset)

    @classmethod
    def register_new_time(cls, attribute_name, time_repr_class):
        if not isinstance(attribute_name, str) or not attribute_name.isidentifier():
            raise ValueError(f"Invalid attribute name ('{attribute_name}') for time representation.")
        if hasattr(cls, attribute_name):
            raise AttributeError(f"Time representation attribute already existing: {attribute_name}.")
        if not hasattr(time_repr_class, "from_time_pair"):
            raise TypeError("Time representation class does not have method from_time_pair.")
        if not hasattr(time_repr_class, "to_time_pair"):
            raise TypeError("Time representation class does not have method to_time_pair.")

        class ModifiedClass(type):
            def __call__(klass, *args, **kwargs):
                time_repr_obj = super().__call__(*args, **kwargs)
                day_frac, utcoffset = time_repr_obj.to_time_pair()
                time_obj = cls(day_frac, utcoffset=utcoffset)
                setattr(time_obj, attribute_name, time_repr_obj)
                return time_obj

        # Create the modified calendar class
        new_class_name = f"{time_repr_class.__name__}In{cls.__name__}"
        modified_time_repr_class = ModifiedClass(new_class_name, (time_repr_class,), {})

        class TimeReprAttribute:
            # This class implements a context dependent attribute
            def __init__(self, attr_name, modified_time_repr_class):
                self.attr_name = attr_name
                self.modified_time_repr_class = modified_time_repr_class

            def __get__(self, instance, owner):
                if instance is None:
                    return self.modified_time_repr_class
                else:
                    assert self.attr_name not in instance.__dict__
                    time_obj = self.modified_time_repr_class.from_time_pair(instance.day_frac, utcoffset=instance.utcoffset)
                    time_repr_obj = getattr(time_obj, self.attr_name)
                    setattr(instance, self.attr_name, time_repr_obj)
                    return time_repr_obj

        setattr(cls, attribute_name, TimeReprAttribute(attribute_name, modified_time_repr_class))


##############################################################################
# Register current time representations
#
Time.register_new_time("western", western.WesternTime)
Time.register_new_time("internet", modern.InternetTime)



import setuptools
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst')) as f:
    long_description = f.read()

# TODO: All setup arguments need to be revised, also check for those to give more info on PyPI page
setuptools.setup(
    name='datetime2',
    version='0.9.2',

    description='New date and time classes',
    long_description=long_description,
    long_description_content_type="text/x-rst",

    url='https://github.com/fricciardi/datetime2',

    author='Francesco Ricciardi',
    author_email='francescor2010@yahoo.it',
    license='BSD License',

    keywords='date time datetime calendar',

    classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Topic :: Scientific/Engineering',
          'Topic :: Software Development :: Libraries :: Python Modules'
          ],

    packages=setuptools.find_packages(exclude=['docs*']),

    platforms=['Platform independent'],

    project_urls={"Documentation": 'https://datetime2.readthedocs.io/en/stable/',
                  "Issue Tracker": "https://github.com/fricciardi/datetime2/issues",
                  "Source": 'https://github.com/fricciardi/datetime2'}
    )



# ISO calendar and Internet time

# Copyright (c) 2013-2022 Francesco Ricciardi
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name(s) of the copyright holders nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AS IS AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

__author__ = "Francesco Ricciardi <francescor2010 at yahoo.it>"


__all__ = ["IsoCalendar"]


import bisect
from fractions import Fraction
from math import floor

from datetime2 import verify_fractional_value

_long_years = frozenset(
    [
        4,
        9,
        15,
        20,
        26,
        32,
        37,
        43,
        48,
        54,
        60,
        65,
        71,
        76,
        82,
        88,
        93,
        99,
        105,
        111,
        116,
        122,
        128,
        133,
        139,
        144,
        150,
        156,
        161,
        167,
        172,
        178,
        184,
        189,
        195,
        201,
        207,
        212,
        218,
        224,
        229,
        235,
        240,
        246,
        252,
        257,
        263,
        268,
        274,
        280,
        285,
        291,
        296,
        303,
        308,
        314,
        320,
        325,
        331,
        336,
        342,
        348,
        353,
        359,
        364,
        370,
        376,
        381,
        387,
        392,
        398,
    ]
)

_weeks_in_previous_years = [0]
for year_index in range(1, 400):
    _weeks_in_previous_years.append(_weeks_in_previous_years[-1] + (52 if year_index not in _long_years else 53))


# This code pretty prints the week-in-previous-years list
# for row in range (20):
#    print('{:3d}: {}'.format(row * 20, " ".join(['{:5d}'.format(_weeks_in_previous_years[y + row * 20]) for y in range(20)])))


##############################################################################
# Iso calendar
#
class IsoCalendar:
    def __init__(self, year, week, day):
        if not isinstance(year, int) or not isinstance(week, int) or not isinstance(day, int):
            raise TypeError("integer argument expected")
        if week < 1 or week > IsoCalendar.weeks_in_year(year):
            raise ValueError(f"Week must be between 1 and number of weeks in year, while it is {week}.")
        if day < 1 or day > 7:
            raise ValueError(f"Day must be between 1 and 7, while it is {day}.")
        self._year = year
        self._week = week
        self._day = day
        self._rata_die = None

    @property
    def year(self):
        return self._year

    @property
    def week(self):
        return self._week

    @property
    def day(self):
        return self._day

    @classmethod
    def from_rata_die(cls, day_count):
        if not isinstance(day_count, int):
            raise TypeError("integer argument expected")
        week_no_less_1, day_less_1 = divmod(day_count - 1, 7)  # ranges: week_no_less_1: free, day_less_1: 0..6
        four_hundred_years, no_of_weeks_in_400 = divmod(week_no_less_1, 20871)  # ranges: four_hundred_years: free, no_of_weeks_in_400: 0..20870
        year_in_400 = bisect.bisect_right(_weeks_in_previous_years, no_of_weeks_in_400)  # range: year_in_400: 1..400
        year = year_in_400 + four_hundred_years * 400
        week = no_of_weeks_in_400 - _weeks_in_previous_years[year_in_400 - 1] + 1
        day = day_less_1 + 1
        iso_day = cls(year, week, day)
        iso_day._rata_die = day_count
        return iso_day

    @staticmethod
    def is_long_year(year):
        return year % 400 in _long_years

    @staticmethod
    def weeks_in_year(year):
        return 52 if year % 400 not in _long_years else 53

    def to_rata_die(self):
        if self._rata_die is None:
            y400, year_in_400 = divmod(self.year - 1, 400)
            self._rata_die = (y400 * 146097 + 7 * (_weeks_in_previous_years[year_in_400] + self.week - 1) + self.day)
        return self._rata_die

    def day_of_year(self):
        return 7 * (self.week - 1) + self.day

    def replace(self, *, year=None, week=None, day=None):
        if year is None:
            year = self.year
        if week is None:
            week = self.week
        if day is None:
            day = self.day
        return type(self)(year, week, day)

    def __repr__(self):
        return f"datetime2.modern.{type(self).__name__}({self.year}, {self.week}, {self.day})"

    def __str__(self):
        if self.year >= 0:
            return f"{self.year:04d}-W{self.week:02d}-{self.day:1d}"
        else:
            return f"{self.year:05d}-W{self.week:02d}-{self.day:1d}"

    name_weekdays = [
        "Monday",
        "Tuesday",
        "Wednesday",
        "Thursday",
        "Friday",
        "Saturday",
        "Sunday",
    ]

    format_functions = {
        "a": lambda self: IsoCalendar.name_weekdays[self.day - 1][:3],
        "A": lambda self: IsoCalendar.name_weekdays[self.day - 1],
        "j": lambda self: f"{self.day_of_year():03d}",
        "w": lambda self: f"{self.day:1d}",
        "W": lambda self: f"{(self.day_of_year() + 7 - self.day) // 7:02d}",
        "y": lambda self: f"{self.year:03d}"[-2:],
        "Y": lambda self: f"{self.year:04d}" if self.year >= 0 else f"-{-self.year:04d}"
    }

    def cformat(self, format_string):
        if not isinstance(format_string, str):
            raise TypeError("Format must be specified with string.")
        output_pieces = []
        for format_chunk in format_string.split("%%"):
            format_parts = format_chunk.split("%")
            chunk_pieces = [format_parts[0]]
            for part in format_parts[1:]:
                if part == "":  # special case: last char is '%'
                    value = "%"
                else:
                    try:
                        value = self.format_functions[part[0]](self)
                    except KeyError:
                        value = "%" + part[0]
                chunk_pieces.append(value)
                chunk_pieces.append(part[1:])
            output_pieces.append("".join(chunk_pieces))
        return "%".join(output_pieces)


##############################################################################
# Internet time representation
#
class InternetTime:
    def __init__(self, beat):
        try:
            beat_fraction = verify_fractional_value(beat, min=0, max_excl=1000)
        except TypeError as exc:
            raise TypeError("beat is not a valid fractional value") from exc
        except ValueError as exc:
            raise ValueError("beat must be equal or greater than 0 and less than 1000.") from exc
        self._beat = beat_fraction

    @property
    def beat(self):
        return self._beat

    @classmethod
    def from_time_pair(cls, day_frac, utcoffset):
        day_frac_valid = verify_fractional_value(day_frac, min=0, max_excl=1, strict=True)
        if utcoffset is None:
            raise TypeError("Internet time can only be used for aware Time instances.")
        utcoffset_valid = verify_fractional_value(utcoffset, min=-1, max=1, strict=True)
        utc_time = day_frac_valid - utcoffset_valid
        beat = (utc_time - floor(utc_time)) * 1000
        internet = cls(beat)
        return internet

    def to_time_pair(self):
        return self._beat / 1000, Fraction(-1, 24)

    def __repr__(self):
        return f"datetime2.modern.{type(self).__name__}({self.beat!r})"

    def __str__(self):
        return f"@{int(self.beat):03d}"

    format_functions = {
        "b": lambda self: f"{int(self.beat):03d}",
        "f": lambda self: f"{int((self.beat - int(self.beat)) * 1000):03d}"
    }

    def cformat(self, format_string):
        if not isinstance(format_string, str):
            raise TypeError("Format must be specified with string.")
        output_pieces = []
        for format_chunk in format_string.split("%%"):
            format_parts = format_chunk.split("%")
            chunk_pieces = [format_parts[0]]
            for part in format_parts[1:]:
                if part == "":  # special case: last char is '%'
                    value = "%"
                else:
                    try:
                        value = self.format_functions[part[0]](self)
                    except KeyError:
                        value = "%" + part[0]
                chunk_pieces.append(value)
                chunk_pieces.append(part[1:])
            output_pieces.append("".join(chunk_pieces))
        return "%".join(output_pieces)



# Gregorian calendar in calendars package

# Copyright (c) 2012-2022 Francesco Ricciardi
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name(s) of the copyright holders nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AS IS AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

__author__ = "Francesco Ricciardi <francescor2010 at yahoo.it>"


import bisect
from fractions import Fraction

from .common import verify_fractional_value


_days_in_month = [
    [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31],
    [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31],
]

_days_in_previous_months = [
    [sum(_days_in_month[leap_year][:month]) for month in range(12)]
    for leap_year in (0, 1)
]


##############################################################################
# Gregorian calendar
#
class GregorianCalendar:
    def __init__(self, year, month, day):
        if not isinstance(year, int) or not isinstance(month, int) or not isinstance(day, int):
            raise TypeError("integer argument expected")
        if month < 1 or month > 12:
            raise ValueError(f"Month must be between 1 and 12, while it is {month}.")
        if day < 1 or day > _days_in_month[GregorianCalendar.is_leap_year(year)][month - 1]:
            raise ValueError(f"Day must be between 1 and number of days in month, while it is {day}.")
        self._year = year
        self._month = month
        self._day = day
        self._rata_die = None

    @property
    def year(self):
        return self._year

    @property
    def month(self):
        return self._month

    @property
    def day(self):
        return self._day

    @classmethod
    def year_day(cls, year, day):
        if not isinstance(year, int) or not isinstance(day, int):
            raise TypeError("integer argument expected")
        if day < 1 or day > (366 if GregorianCalendar.is_leap_year(year) else 365):
            raise ValueError(f"Day must be between 1 and number of days in year, while it is {day}.")
        month = bisect.bisect_left(_days_in_previous_months[GregorianCalendar.is_leap_year(year)], day)
        day_in_month = (day - _days_in_previous_months[GregorianCalendar.is_leap_year(year)][month - 1])
        return cls(year, month, day_in_month)

    @classmethod
    def from_rata_die(cls, day_count):
        if not isinstance(day_count, int):
            raise TypeError("integer argument expected")
        y400, d400 = divmod(day_count - 1, 146097)
        y100, d100 = divmod(d400, 36524)
        y4, d4 = divmod(d100, 1461)
        y1 = d4 // 365
        year_minus_one = (400 * y400 + 100 * y100 + 4 * y4 + y1 - (1 if (y100 == 4 or y1 == 4) else 0))
        days = (day_count - 365 * year_minus_one - year_minus_one // 4 + year_minus_one // 100 - year_minus_one // 400)  # days from january 1st (included) to today
        greg_day = cls.year_day(year_minus_one + 1, days)
        greg_day._rata_die = day_count
        return greg_day

    @staticmethod
    def is_leap_year(year):
        return (year % 4 == 0) and (year % 400 not in (100, 200, 300))

    @staticmethod
    def days_in_year(year):
        return 365 if not GregorianCalendar.is_leap_year(year) else 366

    def to_rata_die(self):
        if self._rata_die is None:
            self._rata_die = (365 * (self._year - 1)
                + (self._year - 1) // 4
                - (self._year - 1) // 100
                + (self._year - 1) // 400
                + (367 * self._month - 362) // 12
                + ((-1 if GregorianCalendar.is_leap_year(self._year) else -2) if self._month > 2 else 0)
                + self._day)
        return self._rata_die

    def weekday(self):
        return (self.to_rata_die() - 1) % 7 + 1

    def day_of_year(self):
        return self.to_rata_die() - (365 * (self._year - 1) + (self._year - 1) // 4
                                     - (self._year - 1) // 100 + (self._year - 1) // 400)

    def replace(self, *, year=None, month=None, day=None):
        if year is None:
            year = self.year
        if month is None:
            month = self.month
        if day is None:
            day = self.day
        return type(self)(year, month, day)

    def __repr__(self):
        return f"datetime2.western.{type(self).__name__}({self.year}, {self.month}, {self.day})"

    def __str__(self):
        if self.year >= 0:
            return f"{self.year:04d}-{self.month:02d}-{self.day:02d}"
        else:
            return f"{self.year:05d}-{self.month:02d}-{self.day:02d}"

    name_weekdays = [
        "Monday",
        "Tuesday",
        "Wednesday",
        "Thursday",
        "Friday",
        "Saturday",
        "Sunday",
    ]
    name_months = [
        "January",
        "February",
        "March",
        "April",
        "May",
        "June",
        "July",
        "August",
        "September",
        "October",
        "November",
        "December",
    ]

    format_functions = {
        "a": lambda self: GregorianCalendar.name_weekdays[self.weekday() - 1][:3],
        "A": lambda self: GregorianCalendar.name_weekdays[self.weekday() - 1],
        "b": lambda self: GregorianCalendar.name_months[self.month - 1][:3],
        "B": lambda self: GregorianCalendar.name_months[self.month - 1],
        "d": lambda self: f"{self.day:02d}",
        "m": lambda self: f"{self.month:02d}",
        "j": lambda self: f"{self.day_of_year():03d}",
        "U": lambda self: f"{(self.day_of_year() + (13 - self.weekday()) % 7) // 7:02d}",
        "w": lambda self: f"{self.weekday():1d}",
        "W": lambda self: f"{(self.day_of_year() + 7 - self.weekday()) // 7:02d}",
        "y": lambda self: f"{self.year:03d}"[-2:],
        "Y": lambda self: f"{self.year:04d}" if self.year >= 0 else f"-{-self.year:04d}",
    }

    def cformat(self, format_string):
        if not isinstance(format_string, str):
            raise TypeError("Format must be specified with string.")
        output_pieces = []
        for format_chunk in format_string.split("%%"):
            format_parts = format_chunk.split("%")
            chunk_pieces = [format_parts[0]]
            for part in format_parts[1:]:
                if part == "":  # special case: last char is '%'
                    value = "%"
                else:
                    try:
                        value = self.format_functions[part[0]](self)
                    except KeyError:
                        value = "%" + part[0]
                chunk_pieces.append(value)
                chunk_pieces.append(part[1:])
            output_pieces.append("".join(chunk_pieces))
        return "%".join(output_pieces)


##############################################################################
# Western time representation
class WesternTime:
    def __init__(self, hour, minute, second, *, timezone=None):
        if not isinstance(hour, int) or not isinstance(minute, int):
            raise TypeError("Hour and minute must be integer")
        if hour < 0 or hour > 23:
            raise ValueError(f"Hour must be between 0 and 23, while it is {hour}.")
        if minute < 0 or minute > 59:
            raise ValueError(f"Minute must be between 0 and 59, while it is {minute}.")
        try:
            second_fraction = verify_fractional_value(second, min=0, max_excl=60)
        except TypeError as exc:
            raise TypeError("Second is not a valid fractional value") from exc
        except ValueError as exc:
            raise ValueError("Second must be equal or greater than 0 and less than 60.") from exc
        self._hour = hour
        self._minute = minute
        self._second = second_fraction
        if timezone is None:
            self._timezone = None
        else:
            try:
                candidate_timezone = verify_fractional_value(timezone, min=-24, max=+24)
            except TypeError as exc:
                raise TypeError("Time zone is not a valid fractional value") from exc
            except ValueError as exc:
                raise ValueError("Time zone must be greater than -24 and less than 24.") from exc
            self._timezone = candidate_timezone

    @property
    def hour(self):
        return self._hour

    @property
    def minute(self):
        return self._minute

    @property
    def second(self):
        return self._second

    @property
    def timezone(self):
        return self._timezone

    @classmethod
    def from_time_pair(cls, day_frac, utcoffset):
        day_frac_valid = verify_fractional_value(day_frac, min=0, max_excl=1, strict=True)
        hour = int(day_frac_valid * 24)
        minute = int((day_frac_valid - Fraction(hour, 24)) * 1440)
        second = (day_frac_valid - Fraction(hour, 24) - Fraction(minute, 1440)) * 86400
        if utcoffset is None:
            western = cls(hour, minute, second)
            return western
        else:
            utcoffset_valid = verify_fractional_value(utcoffset, min=-1, max=1, strict=True)
            western = cls(hour, minute, second, timezone=utcoffset_valid * 24)
            return western

    def to_time_pair(self):
        day_frac = self._second / 86400 + Fraction(self._minute, 1440) + Fraction(self._hour, 24)
        if self._timezone is None:
            return day_frac, None
        else:
            return day_frac, self._timezone / 24

    def replace(self, *, hour=None, minute=None, second=None, timezone=None):
        if hour is None:
            hour = self.hour
        if minute is None:
            minute = self.minute
        if second is None:
            second = self.second
        if timezone is None:
            return type(self)(hour, minute, second, timezone=self.timezone)
        else:
            if self.timezone is None:
                raise TypeError("Can replace timezone only in aware instances.")
            else:
                return type(self)(hour, minute, second, timezone=timezone)

    def __repr__(self):
        if self.timezone is None:
            return f"datetime2.western.{type(self).__name__}({self.hour}, {self.minute}, {self.second!r})"
        else:
            return f"datetime2.western.{type(self).__name__}({self.hour}, {self.minute}, {self.second!r}, timezone={self.timezone!r})"

    def __str__(self):
        time_str = f"{self.hour:02d}:{self.minute:02d}:{int(self.second):02d}"
        if self.timezone is None:
            return time_str
        else:
            tz_hour = int(self.timezone)
            hour_rest = (self.timezone - tz_hour) * 60
            tz_minute = int(hour_rest)
            return f"{time_str}{tz_hour:+02d}:{tz_minute:02d}"

    format_functions = {
        "H": lambda self: f"{self.hour:02d}",
        "I": lambda self: f"{12 if self.hour == 0 else self.hour if self.hour <= 12 else self.hour - 12:02d}",
        "p": lambda self: "AM" if self.hour < 12 else "PM",
        "M": lambda self: f"{self.minute:02d}",
        "S": lambda self: f"{int(self.second):02d}",
        "f": lambda self: f"{int((self.second - int(self.second)) * 1000000):06d}",
    }

    def cformat(self, format_string):
        if not isinstance(format_string, str):
            raise TypeError("Format must be specified with string.")
        output_pieces = []
        for format_chunk in format_string.split("%%"):
            format_parts = format_chunk.split("%")
            chunk_pieces = [format_parts[0]]
            for part in format_parts[1:]:
                if part == "":  # special case: last char is '%'
                    value = "%"
                elif part == 'z':  # this case is too complex to fit in a lambda
                    if self.timezone is None:
                        value = ''
                    else:
                        if self.timezone < 0:
                            tz_sign = '-'
                            abs_timezone = -self.timezone
                        else:
                            tz_sign = '+'
                            abs_timezone = self.timezone
                        tz_hour = int(abs_timezone)
                        remainder_in_minutes = (abs_timezone - tz_hour) * 60
                        tz_minute = int(remainder_in_minutes)
                        remainder_in_seconds = (remainder_in_minutes - tz_minute) * 60
                        tz_second = int(remainder_in_seconds)
                        remainder_in_microseconds = (remainder_in_seconds - tz_second) * 1_000_000
                        tz_microsecond = int(remainder_in_microseconds)
                        if tz_microsecond > 0:
                            value = f"{tz_sign}{tz_hour:02d}:{tz_minute:02d}:{tz_second:02d}.{tz_microsecond:06d}"
                        elif tz_second > 0:
                            value = f"{tz_sign}{tz_hour:02d}:{tz_minute:02d}:{tz_second:02d}"
                        else:
                            value = f"{tz_sign}{tz_hour:02d}:{tz_minute:02d}"
                else:
                    try:
                        value = self.format_functions[part[0]](self)
                    except KeyError:
                        value = "%" + part[0]
                chunk_pieces.append(value)
                chunk_pieces.append(part[1:])
            output_pieces.append("".join(chunk_pieces))
        return "%".join(output_pieces)
